#include <iostream>
using namespace std;

// dimension of linear system
const int n=2;

// L*U 
// Result of factorization to be reused
// for computation of adjoints
// with respect to right-hand side
void LU(double** A) {
  for (int k=0;k<n;k++) {
    for (int i=k+1;i<n;i++) A[i][k]=A[i][k]/A[k][k];
    for (int j=k+1;j<n;j++)
      for (int i=k+1;i<n;i++)
        A[i][j]=A[i][j]-A[i][k]*A[k][j];
  }
}

// L*b
// Forward substitution required for solution
// of linear system
void F(double** A, double* b) {
  for (int i=0;i<n;i++) 
    for (int j=0;j<i;j++)
      b[i]=b[i]-A[i][j]*b[j];
}

// U*b
// Backward substitution required for solution
// of linear system
void B(double** A, double* b) {
  for (int i=n-1;i>=0;i--) {
    for (int j=n-1;j>i;j--)
      b[i]=b[i]-A[i][j]*b[j];
    b[i]=b[i]/A[i][i];
  }
}

// L*b
// Transposed forward substitution required for 
// adjoints with respect to right-hand side
void F_a1s(double** A, double* b) {
  for (int i=0;i<n;i++) {
    for (int j=0;j<i;j++)
      b[i]=b[i]-A[i][j]*b[j];
    b[i]=b[i]/A[i][i];
  }
}

// U*b
// Transposed backward substitution required for 
// adjoints with respect to right-hand side
void B_a1s(double** A, double* b) {
  for (int i=n-1;i>=0;i--) 
    for (int j=n-1;j>i;j--)
      b[i]=b[i]-A[i][j]*b[j];
}

int main() {
// duplication of active data segment
  double *b=new double[n];
  double *b_a1s=new double[n];
  double **A=new double*[n];
  for (int i=0;i<n;i++) A[i]=new double[n];
  double **A_a1s=new double*[n];
  for (int i=0;i<n;i++) A_a1s[i]=new double[n];

//  Jacobian in continuous adjoint mode
  A[0][0]=1./2; A[0][1]=1./3; A[1][0]=1./5; A[1][1]=3./4;
  b[0]=9.; b[1]=11.;
// solution of linear system
  LU(A); F(A,b); B(A,b);

  for (int i=0;i<n;i++) 
    for (int j=0;j<n;j++) A_a1s[i][j]=A[j][i];

// adjoints b_a1s of solution range over
// the Cartesian basis vectors in R^n
// to compute right-hand sides (see Eq. (64))
  for (int i=0;i<n;i++) {
    for (int j=0;j<n;j++) b_a1s[j]=0;
    b_a1s[i]=1;
    F_a1s(A_a1s,b_a1s); B_a1s(A_a1s,b_a1s);
    cout << "dx[" << i << "]/db=( "; 
    for (int j=0;j<n;j++) cout << b_a1s[j] << " ";
    cout << ")" << endl;
// rank-1 adjoint with respect to system matrix
// (see Eq. (65))
    cout << "dx[" << i << "]/dA=( "; 
    for (int ii=0;ii<n;ii++) 
      for (int jj=0;jj<n;jj++) 
        cout << -b_a1s[ii]*b[jj] << " ";
    cout << ")" << endl;
  }

// deallocation of activated data segment
  for (int i=0;i<n;i++) delete [] A[i];
  for (int i=0;i<n;i++) delete [] A_a1s[i];
  delete [] A; delete [] b;
  delete [] A_a1s; delete [] b_a1s;
  return 0;
}
