#include <iostream>
using namespace std;

// dimension of linear system
const int n=2;

// L*U 
// Result of factorization to be reused
// for computation of directional derivatives
// with respect to system matrix and right-hand side
void LU(double** A) {
  for (int k=0;k<n;k++) {
    for (int i=k+1;i<n;i++) A[i][k]=A[i][k]/A[k][k];
    for (int j=k+1;j<n;j++)
      for (int i=k+1;i<n;i++)
        A[i][j]=A[i][j]-A[i][k]*A[k][j];
  }
}

// L*b
// Forward substitution required for solution
// and for directional derivative with respect to
// right-hand side
void F(double** A, double* b) {
  for (int i=0;i<n;i++) 
    for (int j=0;j<i;j++)
      b[i]=b[i]-A[i][j]*b[j];
}

// U*b
// Backward substitution required for solution
// and for directional derivative with respect to
// right-hand side
void B(double** A, double* b) {
  for (int i=n-1;i>=0;i--) {
    for (int j=n-1;j>i;j--)
      b[i]=b[i]-A[i][j]*b[j];
    b[i]=b[i]/A[i][i];
  }
}

int main() {
// duplication of active data segment
  double *b=new double[n];
  double *b_t1s=new double[n];
  double **A=new double*[n];
  for (int i=0;i<n;i++) A[i]=new double[n];
  double **A_t1s=new double*[n];
  for (int i=0;i<n;i++) A_t1s[i]=new double[n];

//  Jacobian in continuous tangent-linear mode
  A[0][0]=1./2; A[0][1]=1./3; A[1][0]=1./5; A[1][1]=3./4;
  b[0]=9.; b[1]=11.;
// solution of linear system
  LU(A); F(A,b); B(A,b);

// directional derivatives A_t1s of A range over
// the Cartesian basis vectors in R^(n x n)
// to compute right-hand sides (see Eq. (38))
  for (int i=0;i<n;i++) 
    for (int j=0;j<n;j++) {
      for (int ii=0;ii<n;ii++) 
        for (int jj=0;jj<n;jj++) A_t1s[ii][jj]=0;
      A_t1s[i][j]=1;
      for (int ii=0;ii<n;ii++) {
        b_t1s[ii]=0;
        for (int jj=0;jj<n;jj++) b_t1s[ii]-=A_t1s[ii][jj]*b[jj];
      }
// existing factorization of A is reused to solve the
// tangent-linear sytem
    F(A,b_t1s); B(A,b_t1s);
    cout << "dx/dA[" << j << "][" << i << "]=( "; 
    for (int ii=0;ii<n;ii++) cout << b_t1s[ii] << " "; 
    cout << ")" << endl;
  }

// directional derivatives of b range over
// the Cartesian basis vectors in R^n
// yielding right-hand sides (see Eq. (38))
  for (int i=0;i<n;i++) {
    for (int j=0;j<n;j++) b_t1s[j]=0;
    b_t1s[i]=1;
// existing factorization of A is reused to solve the
// tangent-linear sytem
    F(A,b_t1s); B(A,b_t1s);
    cout << "dx/db[" << i << "]=( "; 
    for (int j=0;j<n;j++) cout << b_t1s[j] << " "; 
    cout << ")" << endl;
  }

// deallocation of activated data segment
  for (int i=0;i<n;i++) delete [] A[i];
  for (int i=0;i<n;i++) delete [] A_t1s[i];
  delete [] A; delete [] b;
  delete [] A_t1s; delete [] b_t1s;
  return 0;
}
