#include <iostream>
using namespace std;

// declaration of dco's 
// adjoint 1st-order scalar type
#include "dco/a1s/dco_a1s_type.hpp"

// dimension of linear system
const int n=2;

// L*U 
// Factorization is recorded on tape
void LU(dco_a1s_type** A) {
  for (int k=0;k<n;k++) {
    for (int i=k+1;i<n;i++) A[i][k]=A[i][k]/A[k][k];
    for (int j=k+1;j<n;j++)
      for (int i=k+1;i<n;i++)
        A[i][j]=A[i][j]-A[i][k]*A[k][j];
  }
}

// L*b
// Forward substitution is recorded on tape
void F(dco_a1s_type** A, dco_a1s_type* b) {
  for (int i=0;i<n;i++) 
    for (int j=0;j<i;j++)
      b[i]=b[i]-A[i][j]*b[j];
}

// U*b
// Backward substitution is recorded on tape
void B(dco_a1s_type** A, dco_a1s_type* b) {
  for (int i=n-1;i>=0;i--) {
    for (int j=n-1;j>i;j--)
      b[i]=b[i]-A[i][j]*b[j];
    b[i]=b[i]/A[i][i];
  }
}

int main() {
// allocation of active data as of dco's
// adjoint 1st-order scalar type
  dco_a1s_type *b=new dco_a1s_type[n];
  dco_a1s_type **A=new dco_a1s_type*[n];
  for (int i=0;i<n;i++) A[i]=new dco_a1s_type[n];
// tape
  extern dco_a1s_tape_entry dco_a1s_tape[DCO_A1S_TAPE_SIZE];

// Jacobian in discrete adjoint mode
  for (int i=0;i<n;i++) {
    A[0][0]=1./2; A[0][1]=1./3; A[1][0]=1./5; A[1][1]=3./4;
    int va_A[n*n]={A[0][0].va,A[0][1].va,A[1][0].va,A[1][1].va};
    b[0]=9.; b[1]=11.;
    int va_b[n]={b[0].va,b[1].va};
// Solution procedure gets recorded on tape
    LU(A); F(A,b); B(A,b);
// adjoints of solution range over
// the Cartesian basis vectors in R^n
    dco_a1s_tape[b[i].va].a=1;
// tape gets interpreted in current adjoint direction
    dco_a1s_interpret_tape();
    cout << "dx[" << i << "]/db=( "; 
    for (int j=0;j<n;j++) cout << dco_a1s_tape[va_b[j]].a << " "; 
    cout << ")" << endl;
    cout << "dx[" << i << "]/dA=( "; 
    for (int j=0;j<n;j++) 
      for (int k=0;k<n;k++) 
	cout << dco_a1s_tape[va_A[j*n+k]].a << " ";
    cout << ")" << endl; 
    dco_a1s_reset_tape();
  }

// deallocation of active data
  for (int i=0;i<n;i++) delete [] A[i];
  delete [] A; delete [] b;
  return 0;
}
