#include <iostream>
using namespace std;

// declaration of dco's 
// tangent-linear 1st-order scalar type
#include "dco/t1s/dco_t1s_type.hpp"

// dimension of linear system
const int n=2;

// L*U 
// Factorization in tangent-linear mode
void LU(dco_t1s_type** A) {
  for (int k=0;k<n;k++) {
    for (int i=k+1;i<n;i++) A[i][k]=A[i][k]/A[k][k];
    for (int j=k+1;j<n;j++)
      for (int i=k+1;i<n;i++)
        A[i][j]=A[i][j]-A[i][k]*A[k][j];
  }
}

// L*b 
// Forward substitution in tangent-linear mode
void F(dco_t1s_type** A, dco_t1s_type* b) {
  for (int i=0;i<n;i++) 
    for (int j=0;j<i;j++)
      b[i]=b[i]-A[i][j]*b[j];
}

// U*b 
// Backward substitution in tangent-linear mode
void B(dco_t1s_type** A, dco_t1s_type* b) {
  for (int i=n-1;i>=0;i--) {
    for (int j=n-1;j>i;j--)
      b[i]=b[i]-A[i][j]*b[j];
    b[i]=b[i]/A[i][i];
  }
}

int main() {
// allocation of active data as of dco's
// tangent-linear 1st-order scalar type
  dco_t1s_type *b=new dco_t1s_type[n];
  dco_t1s_type **A=new dco_t1s_type*[n];
  for (int i=0;i<n;i++) A[i]=new dco_t1s_type[n];

// Jacobian in discrete tangent-linear mode
//
// directional derivatives of A range over
// the Cartesian basis vectors in R^(n x n)
  for (int i=0;i<n;i++)
    for (int j=0;j<n;j++) {
      A[0][0]=1./2; A[0][1]=1./3; A[1][0]=1./5; A[1][1]=3./4;
      b[0]=9.; b[1]=11.;
      A[i][j].t=1;
      LU(A); F(A,b); B(A,b);
      cout << "dx/dA[" << i << "][" << j << "]=( "; 
      for (int ii=0;ii<n;ii++) cout << b[ii].t << " ";
      cout << ")" << endl;
    }
// directional derivatives of b range over
// the Cartesian basis vectors in R^n
  for (int i=0;i<n;i++) {
    A[0][0]=1./2; A[0][1]=1./3; A[1][0]=1./5; A[1][1]=3./4;
    b[0]=9.; b[1]=11.;
    b[i].t=1;
    LU(A); F(A,b); B(A,b);
    cout << "dx/db[" << i << "]=( "; 
    for (int j=0;j<n;j++) cout << b[j].t << " ";
    cout << ")" << endl;
  }

// deallocation of active data
  for (int i=0;i<n;i++) delete [] A[i];
  delete [] A; delete [] b;
  return 0;
}
